@@ -222,6 +222,7 @@ static bool repack_table_indexes(PGresult *index_details);
 static bool repack_all_indexes(char *errbuf, size_t errsize);
 static void repack_cleanup(bool fatal, const repack_table *table);
 static void repack_cleanup_callback(bool fatal, void *userdata);
+static void repack_cleanup_index(bool fatal, void *userdata);
 static bool rebuild_indexes(const repack_table *table);
 
 static char *getstr(PGresult *res, int row, int col);
@@ -1394,8 +1395,8 @@ repack_one_table(repack_table *table, const char *orderby)
 	command(sql.data, 0, NULL);
 
 	/* While we are still holding an AccessExclusive lock on the table, submit
-	 * the request for an AccessShare lock asynchronously from conn2.
-	 * We want to submit this query in conn2 while connection's
+	 * the request for an ShareUpdateExclusiveLock lock asynchronously from
+	 * conn2. We want to submit this query in conn2 while connection's
 	 * transaction still holds its lock, so that no DDL may sneak in
 	 * between the time that connection commits and conn2 gets its lock.
 	 */
@@ -1411,12 +1412,12 @@ repack_one_table(repack_table *table, const char *orderby)
 
 	/*
 	 * Not using lock_access_share() here since we know that
-	 * it's not possible to obtain the ACCESS SHARE lock right now
+	 * it's not possible to obtain the SHARE UPDATE EXCLUSIVE lock right now
 	 * in conn2, since the primary connection holds ACCESS EXCLUSIVE.
 	 */
-	printfStringInfo(&sql, "LOCK TABLE %s IN ACCESS SHARE MODE",
+	printfStringInfo(&sql, "LOCK TABLE %s IN SHARE UPDATE EXCLUSIVE MODE",
 					 table->target_name);
-	elog(DEBUG2, "LOCK TABLE %s IN ACCESS SHARE MODE", table->target_name);
+	elog(DEBUG2, "LOCK TABLE %s IN SHARE UPDATE EXCLUSIVE MODE", table->target_name);
 	if (PQsetnonblocking(conn2, 1))
 	{
 		elog(WARNING, "Unable to set conn2 nonblocking.");
@@ -1433,7 +1434,7 @@ repack_one_table(repack_table *table, const char *orderby)
 	 * look for and cancel any (potentially dangerous) DDL commands which
 	 * might also be waiting on our table lock at this point --
 	 * it's not safe to let them wait, because they may grab their
-	 * AccessExclusive lock before conn2 gets its AccessShare lock,
+	 * AccessExclusive lock before conn2 gets its ShareUpdateExclusiveLock lock,
 	 * and perform unsafe DDL on the table.
 	 *
 	 * Normally, lock_access_share() would take care of this for us,
@@ -1449,7 +1450,7 @@ repack_one_table(repack_table *table, const char *orderby)
 	}
 
 	/* We're finished killing off any unsafe DDL. COMMIT in our main
-	 * connection, so that conn2 may get its AccessShare lock.
+	 * connection, so that conn2 may get its ShareUpdateExclusiveLock lock.
 	 */
 	command("COMMIT", 0, NULL);
 
@@ -1464,7 +1465,7 @@ repack_one_table(repack_table *table, const char *orderby)
 	 */
 	while ((res = PQgetResult(conn2)))
 	{
-		elog(DEBUG2, "Waiting on ACCESS SHARE lock...");
+		elog(DEBUG2, "Waiting on SHARE UPDATE EXCLUSIVE lock...");
 		if (PQresultStatus(res) != PGRES_COMMAND_OK)
 		{
 			elog(WARNING, "Error with LOCK TABLE: %s", PQerrorMessage(conn2));
@@ -1515,10 +1516,10 @@ repack_one_table(repack_table *table, const char *orderby)
 	 * for the create_table command to go through, so go ahead and obtain
 	 * the lock explicitly.
 	 *
-	 * Since conn2 has been diligently holding its AccessShare lock, it
-	 * is possible that another transaction has been waiting to acquire
-	 * an AccessExclusive lock on the table (e.g. a concurrent ALTER TABLE
-	 * or TRUNCATE which we must not allow). If there are any such
+	 * Since conn2 has been diligently holding its ShareUpdateExclusiveLock
+	 * lock, it is possible that another transaction has been waiting to
+	 * acquire an AccessExclusive lock on the table (e.g. a concurrent ALTER
+	 * TABLE or TRUNCATE which we must not allow). If there are any such
 	 * transactions, lock_access_share() will kill them so that our
 	 * CREATE TABLE ... AS SELECT does not deadlock waiting for an
 	 * AccessShare lock.
@@ -1613,10 +1614,10 @@ repack_one_table(repack_table *table, const char *orderby)
 
 	/*
 	 * 5. Swap: will be done with conn2, since it already holds an
-	 *    AccessShare lock.
+	 *    ShareUpdateExclusiveLock lock.
 	 */
 	elog(DEBUG2, "---- swap ----");
-	/* Bump our existing AccessShare lock to AccessExclusive */
+	/* Bump our existing ShareUpdateExclusive lock to AccessExclusive */
 
 	if (!(lock_exclusive(conn2, utoa(table->target_oid, buffer),
 						 table->lock_table, false)))
@@ -1798,6 +1799,8 @@ lock_access_share(PGconn *conn, Oid relid, const char *target_name)
 	int				i;
 	bool			ret = true;
 
+	pgut_command(conn, "SAVEPOINT repack_sp1", 0, NULL);
+
 	initStringInfo(&sql);
 
 	for (i = 1; ; i++)
@@ -1806,16 +1809,14 @@ lock_access_share(PGconn *conn, Oid relid, const char *target_name)
 		PGresult   *res;
 		int			wait_msec;
 
-		pgut_command(conn, "SAVEPOINT repack_sp1", 0, NULL);
-
 		duration = time(NULL) - start;
 
 		/* Cancel queries unconditionally, i.e. don't bother waiting
 		 * wait_timeout as lock_exclusive() does -- the only queries we
 		 * should be killing are disallowed DDL commands hanging around
 		 * for an AccessExclusive lock, which must be deadlocked at
-		 * this point anyway since conn2 holds its AccessShare lock
-		 * already.
+		 * this point anyway since conn2 holds its ShareUpdateExclusiveLock
+		 * lock already.
 		 */
 		if (duration > (wait_timeout * 2))
 			ret = kill_ddl(conn, relid, true);
@@ -1855,6 +1856,13 @@ lock_access_share(PGconn *conn, Oid relid, const char *target_name)
 	}
 
 	termStringInfo(&sql);
+
+	/*
+	 * Make sure that the current transaction isn't in failed state before
+	 * releasing the savepoint
+	 */
+	if (PQtransactionStatus(conn) == PQTRANS_INTRANS)
+		pgut_command(conn, "RELEASE SAVEPOINT repack_sp1", 0, NULL);
 	pgut_command(conn, "RESET lock_timeout", 0, NULL);
 	return ret;
 }
@@ -1903,9 +1911,10 @@ static bool advisory_lock(PGconn *conn, const char *relid)
  *  conn: connection to use
  *  relid: OID of relation
  *  lock_query: LOCK TABLE ... IN ACCESS EXCLUSIVE query to be executed
- *  start_xact: whether we will issue a BEGIN ourselves. If not, we will
- *              use a SAVEPOINT and ROLLBACK TO SAVEPOINT if our query
- *              times out, to avoid leaving the transaction in error state.
+ *  start_xact: whether we will issue a BEGIN ourselves. Regardless of the
+ *              value we will use a SAVEPOINT and ROLLBACK TO SAVEPOINT if our
+ *              query times out, to avoid leaving the transaction in error
+ *              state.
  */
 static bool
 lock_exclusive(PGconn *conn, const char *relid, const char *lock_query, bool start_xact)
@@ -1914,6 +1923,10 @@ lock_exclusive(PGconn *conn, const char *relid, const char *lock_query, bool sta
 	int			i;
 	bool		ret = true;
 
+	if (start_xact)
+		pgut_command(conn, "BEGIN ISOLATION LEVEL READ COMMITTED", 0, NULL);
+	pgut_command(conn, "SAVEPOINT repack_sp1", 0, NULL);
+
 	for (i = 1; ; i++)
 	{
 		time_t		duration;
@@ -1921,11 +1934,6 @@ lock_exclusive(PGconn *conn, const char *relid, const char *lock_query, bool sta
 		PGresult   *res;
 		int			wait_msec;
 
-		if (start_xact)
-			pgut_command(conn, "BEGIN ISOLATION LEVEL READ COMMITTED", 0, NULL);
-		else
-			pgut_command(conn, "SAVEPOINT repack_sp1", 0, NULL);
-
 		duration = time(NULL) - start;
 		if (duration > wait_timeout)
 		{
@@ -1935,10 +1943,7 @@ lock_exclusive(PGconn *conn, const char *relid, const char *lock_query, bool sta
 				ret = false;
 
 				/* Before exit the loop reset the transaction */
-				if (start_xact)
-					pgut_rollback(conn);
-				else
-					pgut_command(conn, "ROLLBACK TO SAVEPOINT repack_sp1", 0, NULL);
+				pgut_command(conn, "ROLLBACK TO SAVEPOINT repack_sp1", 0, NULL);
 				break;
 			}
 			else
@@ -1981,10 +1986,7 @@ lock_exclusive(PGconn *conn, const char *relid, const char *lock_query, bool sta
 		{
 			/* retry if lock conflicted */
 			CLEARPGRES(res);
-			if (start_xact)
-				pgut_rollback(conn);
-			else
-				pgut_command(conn, "ROLLBACK TO SAVEPOINT repack_sp1", 0, NULL);
+			pgut_command(conn, "ROLLBACK TO SAVEPOINT repack_sp1", 0, NULL);
 			continue;
 		}
 		else
@@ -1997,6 +1999,14 @@ lock_exclusive(PGconn *conn, const char *relid, const char *lock_query, bool sta
 		}
 	}
 
+	/*
+	 * Make sure that the current transaction isn't in failed state before
+	 * releasing the savepoint
+	 */
+	if (PQtransactionStatus(conn) == PQTRANS_INTRANS)
+		pgut_command(conn, "RELEASE SAVEPOINT repack_sp1", 0, NULL);
+	if (!ret && start_xact)
+		pgut_rollback(conn);
 	pgut_command(conn, "RESET lock_timeout", 0, NULL);
 	return ret;
 }
@@ -2078,6 +2088,39 @@ repack_cleanup(bool fatal, const repack_table *table)
 	}
 }
 
+/*
+ * Cleanup temporary objects, created during index repacking (with option
+ * --index-only), on error.
+ */
+static void
+repack_cleanup_index(bool fatal, void *userdata)
+{
+	StringInfoData sql, sql_drop;
+	PGresult   *index_details = (PGresult *) userdata;
+	char	   *schema_name;
+	int			num;
+
+	schema_name = getstr(index_details, 0, 5);
+	num = PQntuples(index_details);
+
+	initStringInfo(&sql);
+	initStringInfo(&sql_drop);
+
+	appendStringInfoString(&sql, "DROP INDEX CONCURRENTLY IF EXISTS ");
+	appendStringInfo(&sql, "\"%s\".",  schema_name);
+
+	for (int i = 0; i < num; i++)
+	{
+		Oid			index = getoid(index_details, i, 1);
+
+		resetStringInfo(&sql_drop);
+		appendStringInfo(&sql_drop, "%s\"index_%u\"", sql.data, index);
+		command(sql_drop.data, 0, NULL);
+	}
+	termStringInfo(&sql_drop);
+	termStringInfo(&sql);
+}
+
 /*
  * Indexes of a table are repacked.
  */
@@ -2086,7 +2129,7 @@ repack_table_indexes(PGresult *index_details)
 {
 	bool				ret = false;
 	PGresult			*res = NULL, *res2 = NULL;
-	StringInfoData		sql, sql_drop;
+	StringInfoData		sql;
 	char				buffer[2][12];
 	const char			*create_idx, *schema_name, *table_name, *params[3];
 	Oid					table, index;
@@ -2117,6 +2160,8 @@ repack_table_indexes(PGresult *index_details)
 		ereport(ERROR, (errcode(EINVAL),
 			errmsg("Unable to obtain advisory lock on \"%s\"", table_name)));
 
+	pgut_atexit_push(repack_cleanup_index, index_details);
+
 	for (i = 0; i < num; i++)
 	{
 		char *isvalid = getstr(index_details, i, 2);
@@ -2241,30 +2286,16 @@ repack_table_indexes(PGresult *index_details)
 	pgut_command(connection, "COMMIT", 0, NULL);
 	ret = true;
 
-drop_idx:
-	resetStringInfo(&sql);
-	initStringInfo(&sql_drop);
-	appendStringInfoString(&sql, "DROP INDEX CONCURRENTLY ");
-	appendStringInfo(&sql, "\"%s\".",  schema_name);
+	pgut_atexit_pop(repack_cleanup_index, index_details);
 
-	for (i = 0; i < num; i++)
-	{
-		index = getoid(index_details, i, 1);
-		if (repacked_indexes[i])
-		{
-			initStringInfo(&sql_drop);
-			appendStringInfo(&sql_drop, "%s\"index_%u\"", sql.data, index);
-			command(sql_drop.data, 0, NULL);
-		}
-		else
-			elog(INFO, "Skipping drop of index_%u", index);
-	}
-	termStringInfo(&sql_drop);
-	termStringInfo(&sql);
+drop_idx:
+	if (num_repacked > 0)
+		repack_cleanup_index(false, index_details);
 
 done:
 	CLEARPGRES(res);
 	free(repacked_indexes);
+	termStringInfo(&sql);
 
 	return ret;
 }
