@@ -9,6 +9,7 @@
 
 #define FRONTEND
 #include "pgut-fe.h"
+#include "common/username.h"
 
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
@@ -16,7 +17,7 @@
 #include <getopt_long.h>
 #endif
 
-char	   *dbname = NULL;
+const char *dbname = NULL;
 char	   *host = NULL;
 char	   *port = NULL;
 char	   *username = NULL;
@@ -33,8 +34,6 @@ worker_conns workers   = {
 
 
 static bool parse_pair(const char buffer[], char key[], char value[]);
-static char *get_username(void);
-
 
 /*
  * Set up worker conns which will be used for concurrent index rebuilds.
@@ -45,7 +44,6 @@ static char *get_username(void);
 void
 setup_workers(int num_workers)
 {
- 	StringInfoData	buf;
  	int i;
  	PGconn *conn;
 
@@ -53,17 +51,23 @@ setup_workers(int num_workers)
 
  	if (num_workers > 1 && num_workers > workers.num_workers)
  	{
- 		initStringInfo(&buf);
- 		if (dbname && dbname[0])
- 			appendStringInfo(&buf, "dbname=%s ", dbname);
- 		if (host && host[0])
- 			appendStringInfo(&buf, "host=%s ", host);
- 		if (port && port[0])
- 			appendStringInfo(&buf, "port=%s ", port);
- 		if (username && username[0])
- 			appendStringInfo(&buf, "user=%s ", username);
- 		if (password && password[0])
- 			appendStringInfo(&buf, "password=%s ", password);
+#define PARAMS_ARRAY_SIZE	6
+
+		const char *keywords[PARAMS_ARRAY_SIZE];
+		const char *values[PARAMS_ARRAY_SIZE];
+
+		keywords[0] = "host";
+		values[0] = host;
+		keywords[1] = "port";
+		values[1] = port;
+		keywords[2] = "user";
+		values[2] = username;
+		keywords[3] = "password";
+		values[3] = password;
+		keywords[4] = "dbname";
+		values[4] = dbname;
+		keywords[5] = NULL;
+		values[5] = NULL;
 
  		if (workers.conns == NULL)
  		{
@@ -87,11 +91,11 @@ setup_workers(int num_workers)
 			 * XXX: could use PQconnectStart() and PQconnectPoll() to
 			 * open these connections in non-blocking manner.
 			 */
- 			conn = PQconnectdb(buf.data);
- 			if (PQstatus(conn) == CONNECTION_OK)
- 			{
- 				workers.conns[i] = conn;
- 			}
+			conn = PQconnectdbParams(keywords, values, true);
+			if (PQstatus(conn) == CONNECTION_OK)
+			{
+				workers.conns[i] = conn;
+			}
 			else
 			{
 				elog(WARNING, "Unable to set up worker conn #%d: %s", i,
@@ -114,10 +118,8 @@ setup_workers(int num_workers)
 		/* In case we bailed out of setting up all workers, record
 		 * how many successful worker conns we actually have.
 		 */
- 		workers.num_workers = i;
-
- 		termStringInfo(&buf);
- 	}
+		workers.num_workers = i;
+	}
 }
 
 /* Disconnect all our worker conns. */
@@ -155,24 +157,12 @@ void disconnect_workers(void)
 void
 reconnect(int elevel)
 {
-	StringInfoData	buf;
 	char		   *new_password;
 
 	disconnect();
-	initStringInfo(&buf);
-	if (dbname && dbname[0])
-		appendStringInfo(&buf, "dbname=%s ", dbname);
-	if (host && host[0])
-		appendStringInfo(&buf, "host=%s ", host);
-	if (port && port[0])
-		appendStringInfo(&buf, "port=%s ", port);
-	if (username && username[0])
-		appendStringInfo(&buf, "user=%s ", username);
-	if (password && password[0])
-		appendStringInfo(&buf, "password=%s ", password);
 
-	connection = pgut_connect(buf.data, prompt_password, elevel);
-	conn2      = pgut_connect(buf.data, prompt_password, elevel);
+	connection = pgut_connect(dbname, host, port, username, password, prompt_password, elevel);
+	conn2      = pgut_connect(dbname, host, port, username, password, prompt_password, elevel);
 
 	/* update password */
 	if (connection)
@@ -185,8 +175,6 @@ reconnect(int elevel)
 			password = pgut_strdup(new_password);
 		}
 	}
-
-	termStringInfo(&buf);
 }
 
 void
@@ -638,39 +626,6 @@ option_find(int c, pgut_option opts1[], pgut_option opts2[])
 	return NULL;	/* not found */
 }
 
-/*
- * Returns the current user name.
- */
-static char *
-get_username(void)
-{
-	char *ret;
-
-#ifndef WIN32
-	struct passwd *pw;
-
-	pw = getpwuid(geteuid());
-	ret = (pw ? pw->pw_name : NULL);
-#else
-	static char username[128];	/* remains after function execute */
-	DWORD		len = sizeof(username) - 1;
-
-	if (GetUserNameA(username, &len))
-		ret = username;
-	else
-	{
-		_dosmaperr(GetLastError());
-		ret = NULL;
-	}
-#endif
-
-	if (ret == NULL)
-		ereport(ERROR,
-			(errcode_errno(),
-			 errmsg("could not get current user name: ")));
-	return ret;
-}
-
 static int
 option_has_arg(char type)
 {
@@ -787,7 +742,7 @@ pgut_getopt(int argc, char **argv, pgut_option options[])
 	(void) (dbname ||
 	(dbname = getenv("PGDATABASE")) ||
 	(dbname = getenv("PGUSER")) ||
-	(dbname = get_username()));
+	(dbname = get_user_name_or_exit(PROGRAM_NAME)));
 
 	return optind;
 }
